/**************************************************

  game.cpp

  The game class for the konnect4 program.
  Gihan Marasingha
  1st January 2000

**************************************************/

#include "grid.h"
#include "game.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <winsock.h>

game::game()
{
  width = 10; height = 10;
  ingame = outg; // not in game
  bell = 0; // don't ring the bell.
  portno = 6587;
  myscore = 0;
  hisscore = 0;
  mytoken = 'X';
  histoken = 'O';
  strcpy(myname,"Prof. Zegarlinski");
  strcpy(hisname,"Prof. Moriaty");
}

void game::swrite(const char *data)
{
  int n;
  n = send(sock,data,strlen(data)+1,0);
  if (n == SOCKET_ERROR)
	 {
		perror("error writing to socket");
		exit(1);
	 }
}

int game::sread(char *data)
{
  int n;
  n = recv(sock,data,256,0);
  if(n == SOCKET_ERROR)
	 {
		perror("error reading from socket");
		exit(1);
	 }
  else if (n==0) // EOF
	 {
		printf("other player has closed the connection.\n");
      end(); // NB make sure this does the right thing.
    }
  return n;
}

void game::command(int i, char *str)
{
  switch(i)
	 {
    case 0: // width command
      setwidth(str); break;
	 case 1: // height command
      setheight(str); break;
    case 2: // connect command
      connect(str); break;
    case 3: // listen command
      listen(); break;
    case 4: // port command
      port(str); break;
    case 5: // score command
      score(); break;
    case 6: // name command
      name(str); break;
	 case 7: case 8: // end or quit commands
		end(); break;
	 case 9: // print command
		print(); break;
	 case 10: // help command
		help(); break;
	 case 11: // rules command
		rules(); break;
	 case 12: // bell command
		setbell(str); break;
	 };
}

void game::setbell(char *arg)
{
  if (arg)
    {
      if (!strcmp(arg,"on"))
		  bell = 1;
		else
	     bell = 0;
    }
  else
    printf("bell is %s.\n",(bell?"on":"off"));
}

void game::setwidth(char *arg)
{
  if(arg)
	 width = atoi(arg);
  else
	 printf("width = %d.\n",width);
}

void game::setheight(char *arg)
{
  if(arg)
	 height = atoi(arg);
  else
	 printf("height = %d.\n",height);
}

void game::connect(char *arg)
{
  struct sockaddr_in servername;
  struct hostent *hostinfo;
  char buffer[256];
  int col;

  if(arg == NULL)
	 {
      printf("Syntax error. Try `help'.\n");
      return;
    }
  sock = socket(PF_INET,SOCK_STREAM,0);
  if(sock == INVALID_SOCKET)
	 {
		perror("socket (client)");
		return;
	 }
  servername.sin_family = AF_INET;
  servername.sin_port = htons(portno);
  hostinfo = gethostbyname(arg);
  if (hostinfo == NULL)
	 {
		printf("Unknown host %s.\n",arg);
		return;
	 }
  servername.sin_addr = *(struct in_addr *) hostinfo->h_addr;

  if (0>::connect(sock,(struct sockaddr *)&servername,
		sizeof(servername)))
	 {
		perror("connect (client)");
		return;
	 }

  sread(buffer); // get data from server.
  strtok(buffer,"\t");
  if(strcmp(buffer,"konnect4"))
	  {
		 printf("The specified host is not a konnect4 server.\n");
		 closesocket(sock);
		 return;
	  }
  printf("Sucessfully connected to %s.\n",arg);
  ingame = ing;
  myscore = 0;
  hisscore = 0;
  strcpy(hisname,strtok(NULL,"\t"));
  printf("Your opponent's name is %s.\n",hisname);
  if(!strcmp(myname,hisname)) // names are the same
	 strcpy(myname,"client"); // change my name to `client'
  swrite(myname);
  strcpy(buffer,strtok(NULL,"\t"));
  width = atoi(buffer);
  strcpy(buffer,strtok(NULL,"\t"));
  height = atoi(buffer);
  printf("height is %d\twidth is %d.\n",height,width);
  mygrid = new grid(width,height);
  mygrid->print();
  col = wait();
  if (col < 0) //  opponent disconnected
	 return;
  mygrid->addtoken(col,histoken);
  mygrid->print();
}

void game::listen(void)
{
  int lsock; // socket used whilst listening.
  int size; // size of clientname
  struct sockaddr_in servername, clientname;
  char buffer[256];

  lsock = socket(AF_INET, SOCK_STREAM, 0);
  if (lsock == INVALID_SOCKET)
	 {
		perror("socket (server)");
		return;
	 }
  servername.sin_family = AF_INET;
  servername.sin_addr.s_addr = htonl(INADDR_ANY);
  servername.sin_port = htons(portno);
  if (bind(lsock, (struct sockaddr *) &servername,
				  sizeof(servername)) < 0)
	 {
		perror("bind");
		closesocket(lsock);
		return;
	 }
  printf("listening...\n");
  if (::listen(lsock,1)<0)
	 {
		perror("listen");
		closesocket(lsock);
		return;
	 }
  sock = accept(lsock, (struct sockaddr *) &clientname, &size);
  closesocket(lsock);
  if (sock == INVALID_SOCKET)
	 {
		perror("accept");
		return;
	 }
  printf("accepted connection from %s.\n",
	 inet_ntoa(clientname.sin_addr));
  ingame = ing;
  myscore = 0;
  hisscore = 0;
  sprintf(buffer,"konnect4\t%s\t%d\t%d",myname,width,height);
  swrite(buffer); // send data to opponent
  sread(hisname); // get opponent's name.
  printf("Your opponent's name is %s.\n",hisname);
  mygrid = new grid(width,height);
  mygrid->print();
  if(bell)
	 putchar('\a');
}

void game::port(char *arg)
{
  if(arg)
    portno = atoi(arg);
  else
    printf("port number = %d.\n",portno);
}

void game::score(void)
{
  printf("Score:\n%d - %s\n%d - %s\n",myscore,
	 myname,hisscore,hisname);
}

void game::name(char *arg)
{
  if(arg)
    strcpy(myname,arg);
  else
    printf("Your name is %s.\n",myname);
}

void game::end(void)
{
  if (ingame == ing)
    {
      score();
      delete mygrid;
		closesocket(sock);
      ingame = outg;
    }
}

void game::print(void)
{
}

void game::help(void)
{
  printf("konnect4 - An internet version of the popular children's game.\n"
	 "by Gihan Marasingha, %s.\n"
	 "Use the following commands to control the program. "
	 "Note that some commands can only be used during a game, "
	 "and some can be used only outside a game. "
	 "variable arguments to commands are shown in capital letters, "
	 "optional arguments are shown within square brackets. If you "
	 "execute a command which can take an optional argument, then "
	 "providing no argument will print out the value of the variable.\n\n"
	 "** listen - runs a server. You must then wait for a connection.\n"
	 "** connect HOSTNAME - connects to the specified Internet host. "
	 "Note that the width and height of the playing grid are determined "
	 "by the server.\n"
	 "** end - ends the current game (does not quit the program).\n"
	 "** quit - ends the current game, if one is running, then exits "
	 "from the program.\n"
	 "** NUMBER - places a game counter in column NUMBER.\n"
	 "** name [NAME] - changes your name to NAME.\n"
	 "** print - prints out the playing grid.\n"
	 "** help - prints out this help info.\n"
	 "** rules - prints out the rules of the game.\n"
	 "** width [NUMBER] - changes the width of the playing grid.\n"
	 "** height [NUMBER] - changes the height of the playing grid.\n"
	 "** port [NUMBER] - change the port address used for Internet "
	 "communications.\n"
	 "** score - prints out the scores of both players.\n"
	 "** bell [on|off] - if bell is on, then a bell rings when "
	 "it's your turn.\n",__DATE__);
}

void game::rules(void)
{
  printf("konnect4 is a game of two players, each player takes turns to "
	 "drop counters into a grid. The winner is the first player to "
	 "form a row, column or diagonal of four consecutive counters.\n");
}

void game::number(int col)
{
  int n;
  char buffer[256];
  
  n = mygrid->addtoken(col,mytoken);
  if (n==mygrid->failure)
    {
      printf("column number out of range.\n");
      return;
    }
  mygrid->print();
  if (n==mygrid->win)
    {
      printf("You have won this round.\n");
      myscore++;
      score();
      mygrid->clear();
		mygrid->print();
    }
  sprintf(buffer,"%d",col);
  swrite(buffer);
  col = wait();
  if(col<0) // opponent had disconnected
    return;
  n = mygrid->addtoken(col,histoken);
  mygrid->print();
  if(n==mygrid->win)
    {
      printf("You have lost this round.\n");
      hisscore++;
      score();
      mygrid->clear();
      mygrid->print();
    }
}

int game::wait(void)
{
  int col;
  char buffer[256];
  printf("waiting for response from %s...\n",hisname);
  if (sread(buffer)<=0)
	 return -1;
  if (bell)
    putchar('\a');
  col = atoi(buffer);
  return col;
}





